/*
* LaGUI: A graphical application framework.
* Copyright (C) 2022-2023 Wu Yiming
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../la_5.h"

extern LA MAIN;
extern struct _tnsMain *T;
extern const char* LA_UDF_EXTENSION_STRINGS[];

int laget_InstanceModified(void* instance);
void laget_TimeString(laTimeInfo *ti, char *result, char** resultdirect);

void laui_DefaultPropDetails(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn *c;
    laUiItem *b;
    c = laFirstColumn(uil);
    laSafeString* s=0;

    if(Extra&&Extra->EndInstance){ //button
        laGeneralUiExtraData* ex=Extra->EndInstance;
        laUiItem* ui=ex->ui;
        if(This&&This->LastPs&&This->EndInstance){
            laProp* p=This->LastPs->p; laPropContainer* pc=p->Container; void* inst=This->LastPs->UseInstance; 
            char FullPath[256]={0}; la_GetPropPackFullPath(This, FullPath);
            strSafePrint(&s,"%s", FullPath); if(ui->ExtraInstructions){ strSafePrint(&s,"(%s)", ui->ExtraInstructions->Ptr); }
            laShowLabel(uil,c,p->Description,0,0)->Flags|=LA_TEXT_LINE_WRAP;
            laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
        }else{
            laShowLabel(uil,c,ui->AT->Description,0,0)->Flags|=LA_TEXT_LINE_WRAP;
            strSafePrint(&s,"%s", ui->AT->Identifier); if(ui->ExtraInstructions){ strSafePrint(&s,"(%s)", ui->ExtraInstructions->Ptr); }
            laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
        }
    }else{
        if(!This || !This->LastPs){ laShowLabel(uil,c,"Can't access property details.",0,0); return; }

        char FullPath[256]={0}; la_GetPropPackFullPath(This, FullPath);
        laProp* p=This->LastPs->p; laPropContainer* pc=p->Container; void* inst=This->LastPs->UseInstance;

        laShowLabel(uil,c,p->Description,0,0)->Flags|=LA_TEXT_LINE_WRAP;
        if(p->PropertyType==(LA_PROP_ENUM)||(LA_PROP_ARRAY|LA_PROP_ENUM)){
            laEnumItem* ei=laGetEnumArrayIndexed(This,This->LastIndex);
            if(ei){ strSafeSet(&s,0);strSafePrint(&s,"Current: %s", ei->Description);laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_UI_FLAGS_HIGHLIGHT|LA_TEXT_LINE_WRAP; }
        }
        laShowLabel(uil,c,FullPath,0,0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;

        if(inst && !pc->OtherAlloc){
            laShowSeparator(uil,c); strSafeSet(&s,0); strSafePrint(&s,"[ %s ]", pc->Name?pc->Name:pc->Identifier, inst);
            int Level=0; laMemNodeHyper* m=memGetHead(inst,&Level);

            laListHandle* users=memGetUserList(inst);
            if(users){ strSafePrint(&s," %d Users", lstCountElements(users)); }
            laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_TEXT_MONO;

            if(Level==2){
                strSafeSet(&s,0);strSafePrint(&s, m->NUID.String);
                if(laget_InstanceModified(inst)){ strSafePrint(&s, " 🌑 Modified"); }else{ strSafePrint(&s, " Clean"); }
                laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_TEXT_MONO;
                laget_TimeString(&m->TimeCreated, FullPath, 0); 
                strSafeSet(&s,0); strSafePrint(&s, "Created at %s", FullPath);
                laShowLabel(uil,c,s->Ptr,0,0)->Flags|=LA_TEXT_MONO;
            }

            if(m){
                if(Level==2){
                    laShowSeparator(uil,c);
                    if(m->FromFile){
                        //strSafeSet(&s,0); strSafePrint(&s,"File:\n%s", m->FromFile->BaseName);
                        //laShowLabel(uil,c,s->Ptr,0,0);
                        char safepath[256]; strEscapePath(safepath, m->FromFile->BaseName->Ptr);
                        strSafeSet(&s,0); strSafePrint(&s,"la.managed_udfs@basename=%s.__self", safepath);
                        laShowItemFull(uil,c,0,s->Ptr,0,0,laui_ManagedUDFItem,0)->Flags|=LA_UI_FLAGS_NO_GAP|LA_UI_FLAGS_NO_DECAL;
                    }else{
                        laShowLabel(uil,c,"<Not saved yet>",0,0)->Flags|=LA_UI_FLAGS_DISABLED;
                    }
                }
            }
            
        }
    }
    strSafeDestroy(&s);
}

void laui_DefaultNodeOperationsPropUiDefine(laUiList *uil, laPropPack *This, laPropPack *OperatorProps, laColumn *UNUSED, int context){
    laColumn *c = laFirstColumn(uil); laColumn* cl,*cr; laSplitColumn(uil,c,0.5); cl=laLeftColumn(c,0);cr=laRightColumn(c,0);
    laShowItemFull(uil, c, This, "base.name",LA_WIDGET_STRING_PLAIN,0,0,0);
    laShowItemFull(uil, cl, This, "base.move",0,"direction=left;text=Move Left;icon=🡰;",0,0);
    laShowItemFull(uil, cr, This, "base.move",0,"text=Move Right;icon=🡲;",0,0);
    laShowItemFull(uil, c, This, "base.delete",0,0,0,0);
}
void laui_DefaultPropUiDefine(laUiList *uil, laPropPack *This, laPropPack *OperatorProps, laColumn *UNUSED, int context){
    laColumn *c;
    laProp *p, *gp;

    if (OperatorProps && OperatorProps->LastPs){
        c = laFirstColumn(uil); gp = OperatorProps->LastPs->p;
        for (p = gp->SubProp->Props.pFirst; p; p = p->Item.pNext){
            if(p->PropertyType!=LA_PROP_OPERATOR) continue;
            laShowItem(uil, c, OperatorProps, p->Identifier);
        }
    }
    if (!This || !This->LastPs) return;

    c = laFirstColumn(uil); gp = This->LastPs->p;
    for (p = gp->SubProp->Props.pFirst; p; p = p->Item.pNext){
        //la_ShowGeneralPropItem(uil, c, This, gp, p, 0, 0, 0);
        laShowItem(uil, c, This, p->Identifier);
    }
}
void laui_StringPropUiDefine(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn *c;
    laProp *p, *gp;

    if (!This || !This->LastPs) return;
    c = laFirstColumn(uil);

    gp = This->LastPs->p;

    if(Extra){
        laShowItem(uil, c, Extra, "copy_selection");
    }else{
        laShowItem(uil, c, This, "copy");
    }
    laShowItem(uil, c, This, "paste");
    laShowItem(uil, c, This, "restore");
    if(gp->Tag&LA_PROP_IS_FILE){ laShowItem(uil, c, This, "get_file_path"); }
    if(gp->Tag&LA_PROP_IS_FOLDER){ laShowItem(uil, c, This, "get_folder_path"); }
}

void laui_IdentifierOnly(laUiList *uil, laPropPack *This, laPropPack *OP_UNUSED, laColumn *Extra, int context){
    laColumn *col = Extra, *c, *cl, *cr, *crl, *crr, *cll, *clr, *clrl, *clrr, *clrrl, *clrrr;
    laUiItem *ui;

    c = laFirstColumn(uil);

    laShowItemFull(uil, c, This, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
}

void laRebuildPathRecursive(laProp *p, char *buf){
    if (p->Container){
        laRebuildPathRecursive(p->Container, buf);
        sprintf(buf, ".%s", p->Identifier);
    }else
        sprintf(buf, "%s", p->Identifier);
}

void laui_IntPropInfo(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *ExtraColumns, int context){
    laColumn *col, *cl, *cr;
    laUiItem *b1,*b2;
    laUiList *tuil;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.2);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);

    laShowItemFull(uil, cr, Base, "base.identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    b1= laOnConditionToggle(uil, cl, 0,0,0,0,0);{
        for(laProp* p=MAIN.ContainerInt->Props.pFirst;p;p=p->Item.pNext){
            laShowItem(uil,cr,Base,p->Identifier);
        }
    }laEndCondition(uil,b1);
}
void laui_FloatPropInfo(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *ExtraColumns, int context){
    laColumn *col, *cl, *cr;
    laUiItem *b1,*b2;
    laUiList *tuil;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.2);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);

    laShowItemFull(uil, cr, Base, "base.identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    b1= laOnConditionToggle(uil, cl, 0,0,0,0,0);{
        for(laProp* p=MAIN.ContainerFloat->Props.pFirst;p;p=p->Item.pNext){
            laShowItem(uil,cr,Base,p->Identifier);
        }
    }laEndCondition(uil,b1);
}

void laui_SubPropInfoDefault(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *crrl, *crrr;
    laUiItem *bracket, *b1, *b2;
    laPropContainer *pc;
    laProp *p;
    char buf[LA_RAW_CSTR_MAX_LEN]={0};

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.2);
    cl = laLeftColumn(c, 1);
    cr = laRightColumn(c, 0);
    laSplitColumn(uil, cr, 0.5);
    crrl = laLeftColumn(cr, 1*6);
    crrr = laRightColumn(cr, 0);

    laPropContainer* subc=Base->LastPs->p->SubProp;
    if (!((laSubProp*)Base->LastPs->p)->GetType){
        if (!subc) subc = la_ContainerLookup(((laSubProp *)Base->LastPs->p)->TargetID);
        if (!subc) return; Base->LastPs->p->SubProp=subc;
    }else{
        subc=((laSubProp*)Base->LastPs->p)->GetType(Base->EndInstance);
    }

    if (!Base->LastPs->p->UDFIsRefer && !Base->LastPs->p->UDFNoCreate && !Base->LastPs->p->UDFIsSingle){
        laShowItem(uil, cr, Base, "identifier")->Flags|=LA_UI_FLAGS_PLAIN;
        bracket = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);
    }
    for (p = subc->Props.pFirst; p; p = p->Item.pNext){
        if (!(p->PropertyType & LA_PROP_SUB)){
            buf[0] = 0;
            sprintf(buf, "la.prop_containers#%s.properties#%s.name", p->Container->Identifier, p->Identifier);
            laShowItemFull(uil, crrl, 0, buf, LA_WIDGET_STRING_PLAIN,0, 0, 0);
            laShowItem(uil, crrr, Base, p->Identifier);
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.name", p->Container->Identifier, p->Identifier); laShowItemFull(uil, crl, 0, buf, 0, laui_SubPropInfoDefault, 0);
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.description", p->Container->Identifier, p->Identifier); laShowItemFull(uil, crr, 0, buf, 0, laui_SubPropInfoDefault, 0);
            laShowSeparator(uil, c);
        }else{
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.name", p->Container->Identifier, p->Identifier); laShowItemFull(uil, crrl, 0, buf, LA_WIDGET_STRING_PLAIN, 0, 0);
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.name", p->Container->Identifier, p->Identifier); laShowItemFull(uil, crl, 0, buf, 0, laui_SubPropInfoDefault, 0);
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.description", p->Container->Identifier, p->Identifier); laShowItemFull(uil, crr, 0, buf, 0, laui_SubPropInfoDefault, 0);
            if (p->UDFIsRefer) laShowLabel(uil, crrr, "🠞", 0, 0);
            elif (p->UDFNoCreate) laShowLabel(uil, crrr, "n", 0, 0);
            else laShowLabel(uil, crrr, "🗇", 0, 0);

            buf[0] = 0;
            sprintf(buf, "la.prop_containers#%s.icon_id", ((laSubProp *)p)->TargetID);
            laShowIcon(uil, cl, 0, buf, 0);

            b1 = laOnConditionToggle(uil, crrl, 0, 0, 0, 0, 0);{ b1->Flags|=LA_TEXT_ALIGN_LEFT;
                sprintf(buf, "text=...%s;", p->Name);
                strSafeSet(&b1->ExtraInstructions, buf);
                laUiDefineFunc f=laui_SubPropInfoDefault;
                if(!context) f=((laSubProp*)p)->GetType?0:(p->UiDefine?p->UiDefine:(p->SubProp&&p->SubProp->UiDefine?p->SubProp->UiDefine:laui_SubPropInfoDefault));
                laShowItemFull(uil, cr, Base, p->Identifier, 0,0, f, context);
            }
            laEndCondition(uil, b1);
            laShowSeparator(uil, c);
        }
    }
    if (!Base->LastPs->p->UDFIsRefer && !Base->LastPs->p->UDFNoCreate && !Base->LastPs->p->UDFIsSingle){
        laEndCondition(uil, bracket);
    }
    //laShowItemFull(uil, crl, Base, "identifier", 0, 0, 0);
    //p = Base->LastPs->p->SubProp->Props.pFirst;
}
void laui_SubPropSelection(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *crrl, *crrr;
    laUiItem *bracket, *b1, *b2;
    laPropContainer *pc;
    laProp *p;
    laPropPack *pp;
    char buf[LA_RAW_CSTR_MAX_LEN]={0};

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.2);
    cl = laLeftColumn(c, 1);
    cr = laRightColumn(c, 0);
    laSplitColumn(uil, cr, 0.8);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 100);
    laSplitColumn(uil, crl, 0.5);
    crrl = laLeftColumn(crl, 150);
    crrr = laRightColumn(crl, 0);

    if (!Base->LastPs->p->SubProp) Base->LastPs->p->SubProp = la_ContainerLookup(((laSubProp *)Base->LastPs->p)->TargetID);

    if (!Base->LastPs->p->UDFIsRefer){
        laShowItemFull(uil, cr, Base, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
        bracket = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);
    }

    for (p = Base->LastPs->p->SubProp->Props.pFirst; p; p = p->Item.pNext){
        if (p->PropertyType == LA_PROP_SUB){
            //buf[0] = 0; sprintf(buf, "la.prop_containers#%s.properties#%s.name", p->Container->Identifier, p->Identifier);
            //laShowItemFull(uil, crrl, 0, buf, LA_WIDGET_STRING_PLAIN, 0, 0);

            buf[0] = 0;
            sprintf(buf, "identifier=%s;", p->Identifier);
            laShowItemFull(uil, crrr, Base, "put_data_block", 0, buf, 0, 0);

            //laShowLabel(uil, crr, "PUT", 0, 0);
            b1 = laOnConditionToggle(uil, crrl, 0, 0, 0, 0, 0);{
                buf[0] = 0; sprintf(buf, "text=...%s;", p->Name); strSafeSet(&b1->ExtraInstructions, buf);
                laShowItemFull(uil, cr, Base, p->Identifier, 0,0, laui_SubPropSelection, 0)->Flags|=LA_TEXT_ALIGN_LEFT;
            }
            laEndCondition(uil, b1);

            laShowSeparator(uil, c);
        }
    }
    if (!Base->LastPs->p->UDFIsRefer){
        laEndCondition(uil, bracket);
    }
    //laShowItemFull(uil, crl, Base, "identifier", 0, 0, 0);
    //p = Base->LastPs->p->SubProp->Props.pFirst;
}

void laui_LinkerSelectionProp(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *crrl, *crrr;
    laUiItem *bracket, *b1, *b2;
    laPropContainer *pc;
    laProp *p;
    laPropPack *pp;
    char buf[LA_RAW_CSTR_MAX_LEN]={0};

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.4);
    cl = laLeftColumn(c, 1);
    cr = laRightColumn(c, 0);

    b1=laBeginRow(uil,cr,0,0);
    laShowItemFull(uil, cr, Base, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    laShowIcon(uil, cr, Base, "pp.last_ps.property.sub_icon_id", 0);
    laEndRow(uil,b1);

    bracket = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);{
        laShowItemFull(uil, cr, Base, "instances", 0,0, laui_LinkerSelectionInstance, 0);
    }
    bracket->State=LA_BT_ACTIVE;
    laEndCondition(uil, bracket);
}
void laui_LinkerSelectionInstance(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *crrl, *crrr;
    laUiItem *bracket, *b1, *b2;
    laPropContainer *pc;
    laProp *p;
    laPropPack *pp;
    char buf[LA_RAW_CSTR_MAX_LEN]={0};

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.4);
    cl = laLeftColumn(c, 1);
    cr = laRightColumn(c, 0);

    laShowItemFull(uil, cr, Base, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);

    b1 = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);{
        laShowItemFull(uil, cr, Base, "children", 0,0, laui_LinkerSelectionProp, 0);
    }
    laEndCondition(uil, b1);
}

void laui_DefaultPanelTitleBar(laUiList *uil, laPropPack *Base, laPropPack *Extra, laUiDefineFunc header){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *cb1, *cb2, *cb3;
    laUiItem *bracket, *b1, *b2;

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.3);
    cl = laLeftColumn(c, 1);
    cr = laRightColumn(c, 0);

    b1=laBeginRow(uil,cr,0,0);
    laUiItem* tu=laShowItemFull(uil, cr, Base, "title", LA_WIDGET_STRING_PLAIN,0, 0, 0);tu->Expand=1;tu->Flags|=LA_UI_FLAGS_NO_TOOLTIP;
    laShowItem(uil, cr, Base, "hide")->Flags|=LA_UI_FLAGS_ICON;
    laShowItem(uil, cr, Base, "dock")->Flags|=LA_UI_FLAGS_ICON;
    laShowItem(uil, cr, Base, "close")->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b1);
    bracket = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);{
        if(!Extra->LastPs->p->SubProp->Props.pFirst){
            laShowLabel(uil, cr, "Panel doesn't have configurable property.", 0, 0);
        }else{
            if(header){
                laShowLabel(uil, cr, "Panel properties:", 0, 0);
                header(uil, Base, Extra, cr, 0);
            }else{
                for(laProp* p = Extra->LastPs->p->SubProp->Props.pFirst;p;p=p->Item.pNext){
                    laShowItem(uil, cr, Extra, p->Identifier);
                }
            }
        }
        //laShowItem(uil, cr, 0,"la.example_int");
        laShowSeparator(uil, c);
        //laShowLabel(uil, c, "Position:", 0, 0);
        //laShowItemFull(uil, c, Base, "position", LA_WIDGET_INT, 0, 0);
        //laShowLabel(uil, c, "Size:", 0, 0);
        //laShowItemFull(uil, c, Base, "size", LA_WIDGET_INT, 0, 0);
        //laShowLabel(uil, c, "Snapping:", 0, 0);
        //laShowItem(uil, c, Base, "snap_enable")->Flags=LA_UI_FLAGS_CYCLE|LA_UI_FLAGS_TRANSPOSE;
        //laShowItemFull(uil, c, Base, "snap", LA_WIDGET_INT, 0, 0);
    }
    laEndCondition(uil, bracket);
}
void laui_DefaultOperatorPanelTitleBar(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr;
    laUiItem *bracket, *b1, *b2;

    c = laFirstColumn(uil);

    b1 = laBeginRow(uil,c,0,0);
    laUiItem* tu=laShowItemFull(uil, c, Base, "title", LA_WIDGET_STRING_PLAIN,0, 0, 0);tu->Expand=1;tu->Flags|=LA_UI_FLAGS_NO_TOOLTIP;
    laShowItem(uil, c, 0, "LA_cancel");//->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil, b1);
}

void laui_ColumnItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *c0, *c, *cl, *cr, *cll, *clr, *crl, *crr, *clll, *cllr, *cll2, *clr2;
    laUiItem *bracket, *b1, *b2;

    c0 = laFirstColumn(uil);
    laSplitColumn(uil, c0, 0.8);
    cll = laLeftColumn(c0, 80);
    c = laRightColumn(c0, 0);
    laSplitColumn(uil, c, 0.7);
    cl = laLeftColumn(c, 0);
    cr = laRightColumn(c, 200);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 70);

    laSplitColumn(uil, cll, 0.5);
    clll = laLeftColumn(cll, 1);
    cllr = laRightColumn(cll, 0);

    laSplitColumn(uil, cl, 0.5);
    cll2 = laLeftColumn(cl, 0);
    clr2 = laRightColumn(cl, 70);

    //laShowItem(uil, c, Base, "self");

    bracket = laOnConditionThat(uil, crr, laPropExpression(0, "live_editor.current_ui")); // laOnConditionThat(uil, 0, "live_editor", la_LiveEditAllowShow);
    laShowItemFull(uil, cll, Base, "select_column", 0, "text=Select;", 0, 0);
    laEndCondition(uil, bracket);

    bracket = laOnConditionThat(uil, c, laPropExpression(Base, "sub_left"));{ //laOnConditionThat(uil, Base, 0, laui_SnapButtonShow); {
        laShowItem(uil, cll2, Base, "split_at");
        laShowItemFull(uil, clr2, Base, "merge_column", 0, "text=Merge;", 0, 0);

        b1 = laOnConditionThat(uil, c, laPropExpression(Base, "snap_state"));{
            laShowItem(uil, crl, Base, "snap_width");
        }
        laElse(uil, b1);{
            laShowLabel(uil, crl, "Align:", 0, 0)->Flags|=LA_TEXT_ALIGN_RIGHT;
        }
        laEndCondition(uil, b1);

        laShowItem(uil, crr, Base, "snap_state");

        laShowSeparator(uil, c0);

        b1 = laOnConditionToggle(uil, clll, 0, 0, 0, 0, 0);{
            laShowItemFull(uil, c, Base, "sub_left", LA_WIDGET_COLLECTION_SINGLE,0, laui_ColumnItem, 0);
        }
        laElse(uil, b1);{
            laShowLabel(uil, cllr, "Left", 0, 0);
        }
        laEndCondition(uil, b1);

        laShowSeparator(uil, c0);

        b1 = laOnConditionToggle(uil, clll, 0, 0, 0, 0, 0);{
            laShowItemFull(uil, c, Base, "sub_right", LA_WIDGET_COLLECTION_SINGLE,0, laui_ColumnItem, 0);
        }
        laElse(uil, b1);{
            laShowLabel(uil, cllr, "Right", 0, 0);
        }
        laEndCondition(uil, b1);
    }
    laElse(uil, bracket);{
        laShowItem(uil, c, Base, "split_column");
    }
    laEndCondition(uil, bracket);
}

void laui_DefaultMenuButtonsFileEntries(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laColumn* c=laFirstColumn(uil);
            
    laShowLabel(uil, c, "Function Test", 0, 0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
    laShowItem(uil, c, 0, "LA_pure_yes_no");
    laShowItem(uil, c, 0, "LA_file_dialog");
    laShowItem(uil, c, 0, "LA_udf_read");
    laShowItemFull(uil, c, 0, "LA_udf_read",0,"mode=append;text=Append",0,0);
    laShowItem(uil, c, 0, "LA_managed_save");
    laShowItem(uil, c, 0, "LA_manage_udf");

    laShowLabel(uil, c, "Other Entries", 0, 0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
    laShowItem(uil, c, 0, "LA_terminate_program");
}
void laui_DefaultMenuButtonsEditEntries(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laColumn* c=laFirstColumn(uil);
    laShowLabel(uil, c, "History", 0, 0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
    laShowItem(uil, c, 0, "LA_undo")->Flags|=LA_UI_FLAGS_NO_CONFIRM;
    laShowItem(uil, c, 0, "LA_redo")->Flags|=LA_UI_FLAGS_NO_CONFIRM;
    laShowSeparator(uil,c);
    laShowItemFull(uil, c, 0, "LA_panel_activator",0,"panel_id=LAUI_histories;text=Histories",0,0);
}
void laui_DefaultMenuButtonsOptionEntries(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laColumn* c=laFirstColumn(uil);
    laShowLabel(uil, c, "Settings", 0, 0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
    laShowItemFull(uil, c, 0, "LA_panel_activator", 0, "panel_id=LAUI_user_preferences;", 0, 0);
    laShowLabel(uil, c, "Information", 0, 0)->Flags|=LA_TEXT_MONO|LA_UI_FLAGS_DISABLED;
    laShowItemFull(uil, c, 0, "LA_panel_activator", 0, "panel_id=LAUI_about;text=About;", 0, 0);
}
void laui_DefaultMenuButtons(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laUiList *muil; laColumn *mc,*c = laFirstColumn(uil);
    muil = laMakeMenuPage(uil, c, "File");{
        mc = laFirstColumn(muil); laui_DefaultMenuButtonsFileEntries(muil,pp,actinst,extracol,0);
    }
    muil = laMakeMenuPage(uil, c, "Edit");{
        mc = laFirstColumn(muil); laui_DefaultMenuButtonsEditEntries(muil,pp,actinst,extracol,0);
    }
    muil = laMakeMenuPage(uil, c, "Options"); {
        mc = laFirstColumn(muil); laui_DefaultMenuButtonsOptionEntries(muil,pp,actinst,extracol,0);
    }
}
void laui_DefaultMenuExtras(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laColumn *c = laFirstColumn(uil);
    laShowLabel(uil, c, MAIN.MenuProgramName, 0, 0)->Expand=1;
}

void laui_DefaultMenuBarActual(laUiList *uil, laPropPack *pp, laPropPack *actinst, laColumn *extracol, int context){
    laUiList *muil;
    laColumn *c,*mc;
    laUiItem* bb;

    c = laFirstColumn(uil);
    laUiItem* row=laBeginRow(uil,c,0,0);

    bb=laOnConditionThat(uil,c,laPropExpression(0,"la.windows.operator_hints"));{

        laShowItem(uil,c,0,"la.windows.operator_hints");

    }laElse(uil,bb);{
        laShowItem(uil, c, 0, "LA_new_panel")->Flags=LA_UI_FLAGS_ICON;

        if(MAIN.MenuButtons){ MAIN.MenuButtons(uil,0,0,0,0); }

        laShowSeparator(uil,c);

        laUiItem* mui=laOnConditionThat(uil,c,laPropExpression(0,"la.windows.maximized_ui"));{
            laShowItemFull(uil,c,0,"LA_canvas_ui_maximize", 0, "icon=🡰;restore=true;text=Restore Layout", 0, 0);
        }laElse(uil,mui);{
            laUiItem* ui=laOnConditionThat(uil,c,laPropExpression(0,"la.windows.maximized_block"));{
                laShowItemFull(uil,c,0,"LA_block_maximize", 0, "icon=🡰;restore=true;text=Restore Layout", 0, 0);
            }laElse(uil,ui);{
                laShowItemFull(uil, c, 0, "LA_switch_layout", 0, "icon=🞐;show_list=true;", 0, 0)->Flags|=LA_UI_FLAGS_ICON;
                laShowItemFull(uil,c,0,"LA_switch_layout", 0, "icon=🡰;reverse=true;", 0, 0)->Flags|=LA_UI_FLAGS_ICON;
                laShowItemFull(uil,c,0,"LA_switch_layout", 0, "icon=🡲;", 0,0)->Flags|=LA_UI_FLAGS_ICON;
                laShowItem(uil, c, 0, "la.windows.layouts.identifier");
                laShowItem(uil,c,0,"LA_new_window")->Flags|=LA_UI_FLAGS_ICON;
            }laEndCondition(uil, ui);
        }laEndCondition(uil,mui);

        laShowSeparator(uil,c);
        laShowItem(uil,c,0,"la.windows.output_color_space");
        laShowItemFull(uil,c,0,"la.windows.output_show_overflow",0,"text=🟩;",0,0);

        if(MAIN.MenuExtras){ laShowSeparator(uil,c); MAIN.MenuExtras(uil,0,0,0,0); }

        laUiItem* b1=laOnConditionThat(uil,c,laPropExpression(0,"la.windows.panels_hidden"));
            laShowSeparator(uil,c);
            muil = laMakeMenuPage(uil, c, "🡻 Minimized");{
                mc = laFirstColumn(muil);
                laShowItemFull(muil, mc, 0, "la.windows.panels_hidden", 0,0, laui_IdentifierOnly,0);
            }
        laEndCondition(uil, b1);
        b1=laOnConditionThat(uil,c,laPropExpression(0,"la.windows.is_fullscreen"));{
            laShowItemFull(uil,c,0,"LA_fullscreen",0,"restore=true;text=Restore;icon=🡷",0,0)->Flags|=LA_UI_FLAGS_EXIT_WHEN_TRIGGERED;
        }laElse(uil,b1);{
            laShowItemFull(uil,c,0,"LA_fullscreen",0,0,0,0)->Flags|=LA_UI_FLAGS_EXIT_WHEN_TRIGGERED;
        }laEndCondition(uil, b1);

    }laEndCondition(uil,bb);

    laEndRow(uil,row);
}
void laui_DefaultMenuBar(laWindow *w){
    laPanel *p;

    p =laCreateTopPanel(w, 0, 0, 0, 0, LA_RH, 0, 0, 0, 0, -1, -1, -1, 0);
    p->IsMenuPanel = 1;
    laui_DefaultMenuBarActual(&p->TitleBar, &p->PP, &p->PropLinkPP, 0, 0);
}

void laui_ThemeListItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr;
    laUiItem *bracket;

    col = laFirstColumn(uil);

    bracket = laBeginRow(uil,col,0,0);
    laShowItemFull(uil, col, Base, "name", LA_WIDGET_STRING_PLAIN,0, 0, 0)->Expand=1;
    laShowItemFull(uil, col, Base, "author", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    laEndRow(uil,bracket);
}
void laui_Theme(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr;
    laUiItem *sui;
    laUiList *tuil;
    laUiItem *ui;

    //col = laFirstColumn(uil);
    //tuil = laMakeGroup(uil, col, "����ժҪ", 0, 0)->Page; {
    //	col = laFirstColumn(tuil);
    //	laSplitColumn(tuil, col, 0.7);
    //	cl = laLeftColumn(col, 0);
    //	cr = laRightColumn(col, 0);
    //	laSplitColumn(tuil, cr, 0.6);
    //	crl = laLeftColumn(cr, 0);
    //	crr = laRightColumn(cr, 70);
    //	laShowItemFull(tuil, cl, 0, "themes.name", 0, 0, 0);
    //	laShowItemFull(tuil, crl, 0, "themes.author", 0, 0, 0);
    //	laShowLabel(tuil, crr, 0, "EXEC", 0, 0, 0);
    //}

    col = laFirstColumn(uil); laSplitColumn(uil,col,0.5);
    cl=laLeftColumn(col,0);
    cr=laRightColumn(col,0);

    ui=laBeginRow(uil,col, 0,0);
    laShowItem(uil,col,Base, "name")->Expand=1;
    laShowItem(uil,col,Base, "delete");
    laEndRow(uil, ui);

    laShowLabel(uil, col, "Basics:",0,0);
    laShowLabel(uil, cl, "Base Color:",0,0)->Flags|=LA_TEXT_ALIGN_RIGHT; laShowItem(uil, cr, Base, "color");
    laShowLabel(uil, cl, "Accent Color:",0,0)->Flags|=LA_TEXT_ALIGN_RIGHT; laShowItem(uil, cr, Base, "accent_color");
    laShowItem(uil, cr, Base, "cursor_alpha");
    laShowItem(uil, cr, Base, "selection_alpha");
    laShowItem(uil, cr, Base, "inactive_mix");
    laShowItem(uil, cr, Base, "inactive_saturation");

    laShowLabel(uil, col, "Nodes:",0,0);
    laShowItem(uil, cr, Base, "wire_brightness");
    laShowItem(uil, cr, Base, "wire_saturation");
    laShowItem(uil, cr, Base, "wire_transparency");

    laShowLabel(uil, col, "Meshes:",0,0);
    laShowItem(uil, cl, Base, "edge_transparency");
    laShowItem(uil, cl, Base, "edge_brightness");
    laShowItem(uil, cr, Base, "vertex_transparency");
    laShowItem(uil, cr, Base, "vertex_brightness");
    laShowItem(uil, col, Base, "svertex_transparency");
    laShowItem(uil, col, Base, "sedge_transparency");
    laShowItem(uil, col, Base, "sface_transparency");    

    laShowLabel(uil, col, "Widgets:",0,0);
    laShowItem(uil, col, Base, "boxed_themes");
}
void laui_BoxedThemeItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl,*crr;
    laUiItem *bracket;
    laUiList *tuil;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.8);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);

    laShowItemFull(uil, cr, Base, "name", LA_WIDGET_STRING_PLAIN, 0 ,0,0);

    bracket = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);{
        laShowItem(uil, cr, Base, "normal");
        laShowItem(uil, cr, Base, "active");
        laShowItem(uil, cr, Base, "border");
        laShowItem(uil, cr, Base, "text");
        laShowItem(uil, cr, Base, "text_active");
        laShowItem(uil, cr, Base, "alpha");
        laShowLabel(uil,crl,"Margins:",0,0);
        laShowItem(uil, crl, Base, "margins")->Flags=LA_UI_FLAGS_TRANSPOSE;
        laShowLabel(uil,crr,"Paddings:",0,0);
        laShowItem(uil, crr, Base, "paddings")->Flags=LA_UI_FLAGS_TRANSPOSE;
    }
    laEndCondition(uil, bracket);
}

void laui_PropertyContainerList(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl, *crr, *cll, *clr, *crrl, *crrr;
    laUiItem *bracket;
    laUiList *tuil;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.5);
    cl = laLeftColumn(col, 60);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.25);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);
    laSplitColumn(uil, crr, 0.3);
    crrl = laLeftColumn(crr, 0);
    crrr = laRightColumn(crr, 0);

    laShowLabel(uil, cl, "Select", 0, 0);
    laShowLabel(uil, crl, "Identifier", 0, 0);
    laShowLabel(uil, crrl, "Name", 0, 0);
    laShowLabel(uil, crrr, "Description", 0, 0);
    laShowColumnAdjuster(uil, col);

    laShowItem(uil, col, Base, "properties");
}
void laui_WindowListItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr;
    laUiItem *bracket;
    laUiList *tuil;

    cl = laFirstColumn(uil);
    laSplitColumn(uil, cl, 0.2);
    cll = laLeftColumn(cl, 1);
    clr = laRightColumn(cl, 0);

    laShowItem(uil, clr, Base, "title");

    bracket = laOnConditionToggle(uil, cll, 0, 0, 0, 0, 0);{
        laShowLabel(uil, clr, "Layouts", 0, 0);
        laShowItem(uil, clr, Base, "layouts");
        laShowLabel(uil, clr, "TopPanels", 0, 0);
        laShowItem(uil, clr, Base, "panels");
    }
    laEndCondition(uil, bracket);
}
void laui_LayoutListItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr;
    laUiItem *bracket;
    laUiList *tuil;

    cl = laFirstColumn(uil);
    laSplitColumn(uil, cl, 0.2);
    cll = laLeftColumn(cl, 1);
    clr = laRightColumn(cl, 0);

    laShowItem(uil, clr, Base, "title");

    bracket = laOnConditionToggle(uil, cll, 0, 0, 0, 0, 0);{
        laShowItem(uil, clr, Base, "panels");
    }
    laEndCondition(uil, bracket);
}
void laui_PanelListItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr;
    laUiItem *bracket;
    laUiList *tuil;

    cl = laFirstColumn(uil);
    laSplitColumn(uil, cl, 0.2);
    cll = laLeftColumn(cl, 1);
    clr = laRightColumn(cl, 0);

    laShowItem(uil, clr, Base, "title");

    bracket = laOnConditionToggle(uil, cll, 0, 0, 0, 0, 0);{

        laShowItemFull(uil, clr, Base, "size", LA_WIDGET_INT,0, 0, 0);
    }
    laEndCondition(uil, bracket);
}
void laui_UiTemplateListItemLiveEdit(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr;
    laUiItem *bracket;
    laUiList *tuil;

    cl = laFirstColumn(uil);

    //bracket = laOnConditionThat(uil, cl, laNot(laPropExpression(Base, "for_panel"))); {

    bracket = laOnConditionThat(uil, cl, laAnd(laNot(laPropExpression(Base, "for_panel")), laOr(laNot(laPropExpression(Base, "target")), laEqual(laPropExpression(0, "live_editor.current_ui.prop_pack.last_ps.property.sub"), laPropExpression(Base, "target")))));{
        laShowItemFull(uil, cl, Base, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    }
    laElse(uil, bracket);{
        laShowItemFull(uil, cl, Base, "identifier", LA_WIDGET_STRING_PLAIN, "disabled=true;",0, 0);
    }
    laEndCondition(uil, bracket);
}
void laui_UiTemplateListItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *cll, *clr, *clll, *cllr, *clrl, *clrr, *clrrl, *clrrr;
    laUiItem *bracket, *b1;
    laUiList *tuil;

    cl = laFirstColumn(uil);
    laSplitColumn(uil, cl, 0.2);
    cll = laLeftColumn(cl, LA_2RH);
    clr = laRightColumn(cl, 0);

    laSplitColumn(uil, cll, 0.5);
    clll = laLeftColumn(cll, 0);
    cllr = laRightColumn(cll, 0);

    laSplitColumn(uil, clr, 0.5);
    clrl = laLeftColumn(clr, 0);
    clrr = laRightColumn(clr, 200);

    laSplitColumn(uil, clrr, 0.2);
    clrrl = laLeftColumn(clrr, 1);
    clrrr = laRightColumn(clrr, 0);

    bracket = laOnConditionThat(uil, cll, laPropExpression(Base, "define_func"));{
        laShowLabel(uil, clll, "f", 0, 0);
    }
    laEndCondition(uil, bracket);

    bracket = laOnConditionThat(uil, cll, laPropExpression(Base, "descriptor"));{
        laShowLabel(uil, cllr, "🠞", 0, 0);
    }
    laEndCondition(uil, bracket);

    laShowItemFull(uil, clrl, Base, "identifier", LA_WIDGET_STRING_PLAIN,0, 0, 0);

    bracket = laOnConditionThat(uil, cll, laPropExpression(Base, "target_id"));{
        laShowIcon(uil, clrrl, Base, "target.icon_id", 0);
        laShowItemFull(uil, clrrr, Base, "target_id", LA_WIDGET_STRING_PLAIN,0, 0, 0);
    }
    laElse(uil, bracket);{
        b1 = laOnConditionThat(uil, cll, laPropExpression(Base, "for_panel"));{
            //laShowLabel(uil, clrrl, 0,, 0, 0);
            laShowLabel(uil, clrrr, "PANEL", 0, 0);
        }
        laElse(uil, b1);{
            laShowLabel(uil, clrrr, "-", 0, LA_WIDGET_STRING_PLAIN);
        }
        laEndCondition(uil, b1);
    }
    laEndCondition(uil, bracket);
}

void tnsui_GroupItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl, *crr;
    laUiItem *b1;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.3);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);

    laShowItem(uil, cr, Base, "name");
    b1 = laOnConditionToggle(uil, cl, 0, 0, 0, 0, 0);{
        laShowItem(uil, cr, Base, "objects");
    }
    laEndCondition(uil, b1);
}
void tnsui_ObjectGroupItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl, *crr;
    laUiItem *b1;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.3);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);

    laShowItem(uil, cr, Base, "group.name");
}
void tnsui_GroupObjectItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl, *crr;
    laUiItem *b1;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.3);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);

    laShowItem(uil, cr, Base, "object.identifier");
}
void tnsui_MaterialItem(laUiList *uil, laPropPack *Base, laPropPack *UNUSED_This, laColumn *UNUSED_Colums, int context){
    laColumn *col, *cl, *cr, *crl, *crr;
    laUiItem *b1;

    col = laFirstColumn(uil);
    laSplitColumn(uil, col, 0.3);
    cl = laLeftColumn(col, 1);
    cr = laRightColumn(col, 0);
    laSplitColumn(uil, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);

    laShowItem(uil, cr, Base, "object.identifier");
}

void laui_PlainSceneListItem(laUiList *uil, laPropPack *This, laPropPack *OP_UNUSED, laColumn *Extra, int context){
    laColumn *col = Extra, *c, *cl, *cr, *crl, *crr, *cll, *clr, *clrl, *clrr, *clrrl, *clrrr;
    laUiItem *ui;

    //if (Extra)c = Extra;

    c = laFirstColumn(uil);

    laShowItemFull(uil, c, This, "name", LA_WIDGET_STRING_PLAIN, 0, 0, 0);
}

void laui_DataRestorePage(laUiList *uil, laPropPack *THIS_UNUSED, laPropPack *Operator, laColumn *UNUSED, int context){
    laColumn *col = laFirstColumn(uil), *c, *cl, *cr, *crl, *crr, *cll, *clr, *clrl, *clrr, *clrrl, *clrrr;
    laUiList *u;
    laUiItem *b;

    laShowItemFull(uil, col, 0, "la", 0, 0, laui_SubPropSelection, 0);
    laShowItemFull(uil, col, 0, "tns", 0, 0, laui_SubPropSelection, 0);
}
void laui_FileBrowserFileItem(laUiList *uil, laPropPack *This, laPropPack *OP_UNUSED, laColumn *Extra, int context){
    laColumn *c = Extra, *cl, *cr, *crl, *crr, *cll, *clr, *clrl, *clrr, *clrrl, *clrrr;
    laUiItem *ui;

    laSplitColumn(0, c, 0.1);
    cll = laLeftColumn(c, 1);
    clr = laRightColumn(c, 0);
    laSplitColumn(0, clr, 0.7);
    clrl = laLeftColumn(clr, 0);
    clrr = laRightColumn(clr, 0);
    laSplitColumn(0, clrr, 0.5);
    clrrl = laLeftColumn(clrr, 0);
    clrrr = laRightColumn(clrr, 0);

    laShowItemFull(uil, cll, This, "type", LA_WIDGET_ENUM_ICON_PLAIN, 0, 0, 0)->Flags|=LA_TEXT_ALIGN_RIGHT;
    //laShowItem(uil, cll, This, "is_folder");
    laShowItemFull(uil, clrl, This, "name", LA_WIDGET_STRING_PLAIN, 0, 0, 0);
    ui = laOnConditionThat(uil, clrr, laNot(laPropExpression(This, "is_folder")));{
        laShowItemFull(uil, clrrl, This, "time_modified.time_string", LA_WIDGET_STRING_PLAIN, 0, 0, 0);
        laShowItem(uil, clrrr, This, "size")->Flags|=LA_UI_FLAGS_NO_DECAL;
    }
    laEndCondition(uil, ui);
    //laShowSeparator(uil,cll)
    //laShowItemFull(uil, clrrl, This, "time_modified.time_string", LA_WIDGET_STRING_PLAIN, 0, 0);
    //laShowItem(uil, clrrr, This, "size");
}
void laui_FileBrowserDiskItem(laUiList *uil, laPropPack *This, laPropPack *OP_UNUSED, laColumn *Extra, int context){
    laColumn *col = Extra, *c, *cl, *cr, *crl, *crr;
    laUiItem *ui;

    c = Extra;
    laSplitColumn(0, c, 0.2);
    cl = laLeftColumn(c, LA_2RH);
    cr = laRightColumn(c, 0);
    laSplitColumn(0, cr, 0.5);
    crl = laLeftColumn(cr, 0);
    crr = laRightColumn(cr, 0);
    //laShowColumnAdjuster(uil, c);

    laShowItemFull(uil, cl, This, "id", LA_WIDGET_STRING_PLAIN, 0, 0, 0);
    laShowItem(uil, crl, This, "total_gb");
    laShowItem(uil, crr, This, "free_gb");
}
void laui_FileBrowserFileList(laUiList *uil, laPropPack *THIS_UNUSED, laPropPack *Operator, laColumn *UNUSED, int context){
    laColumn *col = laFirstColumn(uil), *c, *cl, *cr, *crl, *crr, *cll, *clr, *clrl, *clrr, *clrrl, *clrrr, *ulc;
    laUiList *u; laUiItem *b; laUiList *ul;

    laSplitColumn(uil, col, 0.25);
    cl = laLeftColumn(col, 10);
    cr = laRightColumn(col, 0);

    laShowItemFull(uil, cl, Operator, "folder_up", LA_WIDGET_BUTTON_NO_CONFIRM, 0 ,0,0);
    laShowItem(uil, cr, Operator, "path")->Expand=1;

    laUiItem* r=laBeginRow(uil,cr,0,0);
    laShowItem(uil, cr, Operator, "file_name")->Expand=1;
    laShowItem(uil, cr, Operator, "confirm")->Flags|=LA_UI_FLAGS_HIGHLIGHT|LA_UI_FLAGS_NO_CONFIRM;
    laEndRow(uil,r);

    laShowSeparator(uil,col);

    laUiItem* left=laMakeEmptyGroup(uil, cl, "left", 0); ul=left->Page; left->Flags|=LA_UI_FLAGS_NO_DECAL; ul->HeightCoeff=-1;
    ulc=laFirstColumn(ul);

    b=laOnConditionThat(ul,ulc,laPropExpression(Operator,"use_type"));{
        u = laMakeGroup(ul, ulc, "Use Format", 0)->Page; c = laFirstColumn(u);
        laShowItemFull(u,c,Operator,"available_extensions",0,0,laui_IdentifierOnly,0)->Flags|=LA_UI_FLAGS_NO_DECAL;
    }laEndCondition(ul,b);

    u = laMakeFoldableGroup(ul, ulc, "Bookmarks", 0, 0, 0)->Page; c = laFirstColumn(u);
    laShowItemFull(u,c,Operator,"bookmarks",0,0,laui_IdentifierOnly,0)->Flags|=LA_UI_FLAGS_NO_DECAL;

    u = laMakeGroup(ul, ulc, "Logic Drives", 0)->Page;
    c = laFirstColumn(u); laSplitColumn(u, c, 0.2); cll = laLeftColumn(c, LA_2RH); clr = laRightColumn(c, 0);
    laSplitColumn(u, clr, 0.5); clrl = laLeftColumn(clr, 0); clrr = laRightColumn(clr, 0);
    laShowColumnAdjuster(u, c);
    laShowItem(u, c, Operator, "disk_list");

    b = laMakeGroup(uil, cr, "File List", 0);b->State=LA_UI_ACTIVE; u=b->Page;
    u->HeightCoeff = -1;
    c = laFirstColumn(u); laSplitColumn(u, c, 0.1); cll = laLeftColumn(c, 1); clr = laRightColumn(c, 0);
    laSplitColumn(u, clr, 0.7); clrl = laLeftColumn(clr, 0); clrr = laRightColumn(clr, 0);
    laSplitColumn(u, clrr, 0.5); clrrl = laLeftColumn(clrr, 0); clrrr = laRightColumn(clrr, 0);
    laShowColumnAdjuster(u, c);
    laShowLabel(u, clrl, "File Name", 0, 0);
    laShowLabel(u, clrrl, "Last Modified On", 0, 0);
    laShowLabel(u, clrrr, "File Size", 0, 0);
    laShowItem(u, c, Operator, "file_list")->Flags|=LA_UI_FLAGS_NO_DECAL;
}
void laui_LinkerPanel(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *cll, *clr, *crl, *crr, *crrl, *crrr, *gc;
    laUiList *gu;
    laUiItem *bracket, *b1, *b2;
    laPropContainer *pc;
    laProp *p;
    laPropPack *pp;
    char buf[LA_RAW_CSTR_MAX_LEN]={0};

    c = laFirstColumn(uil);
    laSplitColumn(uil, c, 0.2);
    cl = laLeftColumn(c, 100);
    cr = laRightColumn(c, 0);
    laSplitColumn(uil, cr, 0.2);
    crl = laLeftColumn(cr, 1);
    crr = laRightColumn(cr, 0);

    gu = laMakeGroup(uil, c, " UDF File Content", 0)->Page;{
        gc = laFirstColumn(gu);
        laShowItem(gu, gc, OperatorInst, "root_nodes");
    }
}

void laui_ManagedUDFItem(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c=laFirstColumn(uil);
    laSplitColumn(uil,c,0.5);
    laColumn *cl=laLeftColumn(c, 1);
    laColumn *cr=laRightColumn(c,0);
    laUiItem* b2=laOnConditionThat(uil,cr,laPropExpression(Base,"udf"));{
        laUiItem* r=laBeginRow(uil,c,0,0);
        laUiItem* b=laOnConditionToggle(uil,c,0,0,0,0,0);strSafeSet(&b->ExtraInstructions,"text=🛈;");b->Flags|=LA_UI_FLAGS_PLAIN;{
            laShowItemFull(uil,c,Base,"udf.path",LA_WIDGET_STRING_PLAIN, 0 ,0,0);
        }laElse(uil,b);{
            laShowItemFull(uil,c,Base,"basename",LA_WIDGET_STRING_PLAIN, 0 ,0,0);
        }laEndCondition(uil,b);
        laEndRow(uil, r);
    }laElse(uil,b2);{
        laShowItemFull(uil,c,Base,"basename",LA_WIDGET_STRING_PLAIN, 0 ,0,0);
    }laEndCondition(uil,b2);
}
void laui_ManagedUDFOps(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c=laFirstColumn(uil);
    laSplitColumn(uil,c,0.3);
    laColumn *cl=laLeftColumn(c, 1);
    laColumn *cr=laRightColumn(c,0);
    laUiItem* b1=laOnConditionThat(uil, cr, laPropExpression(Base, "udf"));{
        laShowItem(uil,cl,Base,"udf.modified");
        laUiItem* b=laBeginRow(uil,cr,0,0);
        laUiItem* b2=laOnConditionToggle(uil,cr,0,0,0,0,0);strSafeSet(&b2->ExtraInstructions,"text=🛈;");b2->Flags|=LA_UI_FLAGS_PLAIN;{
            laUiItem* ui=laShowItemFull(uil,cr,Base,"udf.path",LA_WIDGET_STRING_PLAIN, 0 ,0,0);ui->Expand=1;strSafeSet(&ui->ExtraInstructions,"text=Change");
            laShowItem(uil,cr,&ui->PP,"get_file_path");
        }laElse(uil,b2);{
            laShowItemFull(uil,cr,Base,"basename", LA_WIDGET_STRING_PLAIN, 0 ,0,0)->Expand=1;
        }laEndCondition(uil,b2);
        laEndRow(uil, b);
    }laElse(uil,b1);{
        laShowItemFull(uil,cr,0,"LA_managed_save_new_file",0,"feedback=NONE;",0,0);
    }laEndCondition(uil,b1);
}
void laui_ManagedPropInstance(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    if(!Base ||! Base->EndInstance ||Base->LastPs->p->PropertyType!=LA_PROP_SUB) return;
    laColumn *c=laFirstColumn(uil);
    laSplitColumn(uil,c,0.3); laColumn *cl=laLeftColumn(c, 1); laColumn *cr=laRightColumn(c,0);
    laSplitColumn(uil,cr,0.6); laColumn *crl=laLeftColumn(cr, 0); laColumn *crr=laRightColumn(cr,30);

    laPropContainer* pc=la_EnsureSubTarget(Base->LastPs->p, Base->EndInstance);

    if(pc->UDFPropagate){
        char buf[128]; sprintf(buf,"Assign all \"%s\" into:",pc->Name);
        laShowLabel(uil,crl,buf,0,0)->Flags|=LA_TEXT_ALIGN_RIGHT|LA_UI_FLAGS_DISABLED;
        laShowItem(uil,crr,Base,"__single_udf_propagate");
    }
    
    if(pc->OtherAlloc || Base->LastPs->p->UDFIsSingle){
        //laShowLabel(uil,cr,"(Item not allocated by memAcquire)",0,0)->Expand=1;
    }else{
        if(pc->Hyper==2){
            laShowItem(uil,cl,Base,"__modified"); laSplitColumn(uil,crl,0.4); laColumn *crll=laLeftColumn(crl, 0); laColumn *crlr=laRightColumn(crl,15);
            laUiItem* idui=laShowItem(uil,crll,Base,"identifier");idui->Flags|=LA_UI_FLAGS_PLAIN;idui->Expand=1;
            laShowItem(uil,crlr,Base,"__uid")->Flags|=LA_UI_FLAGS_NO_DECAL;
            if(pc->UDFPropagate){
                laColumn *crrl,*crrr,*udfc,*optionc; laSplitColumn(uil,crr,0.7); crrl=laLeftColumn(crr,0); crrr=laRightColumn(crr,1);
                laSplitColumn(uil,crrl,0.3); udfc=laLeftColumn(crrl,0); optionc=laRightColumn(crrl,8);
                laUiItem* uc=laOnConditionToggle(uil,crrr,0,0,0,0,0);{ strSafeSet(&uc->ExtraInstructions,"text=⮯;");
                    laShowItem(uil,udfc,Base,"__file");
                    laUiItem* sr=laBeginRow(uil,optionc,1,0);
                    laShowItemFull(uil,optionc,Base,"__udf_propagate",0,"force=true;text=Force;",0,0)->Flags|=LA_TEXT_ALIGN_CENTER;
                    laShowItemFull(uil,optionc,Base,"__udf_propagate",0,0,0,0)->Flags|=LA_TEXT_ALIGN_CENTER;
                    laEndRow(uil,sr);
                }laElse(uil,uc);{
                    laShowItem(uil,crrl,Base,"__file");
                }laEndCondition(uil,uc);
            }else{
                laShowItem(uil,crr,Base,"__file");
            }
        }else{
            //laShowItem(uil,crl,Base,"identifier")->Flags|=LA_UI_FLAGS_PLAIN;
        }
    }
    for(laProp*p=pc->Props.pFirst;p;p=p->Item.pNext){
        if(p->PropertyType!=LA_PROP_SUB || p->UDFIsRefer) continue;
        la_EnsureSubTarget(p,0); 
        if(p->SubProp &&p->SubProp->Hyper!=2 &&(!p->UDFIsSingle)) continue;
        if(p->UDFNoCreate || p->UDFHideInSave) continue;
        if(p==pc->SaverDummy){
            laShowItem(uil,cl,Base,"__single_saver_dummy.__modified");
            laShowItem(uil,crl,Base,"identifier")->Flags|=LA_UI_FLAGS_PLAIN;
            laShowItem(uil,crr,Base,"__single_saver_dummy.__file");
            continue;
        }
        laUiItem* b=laOnConditionToggle(uil,c,0,0,0,0,0);{ strSafePrint(&b->ExtraInstructions,"text=⮡ %s;",p->Identifier);
            b->Flags|=LA_UI_FLAGS_NO_DECAL|LA_TEXT_ALIGN_LEFT; b->State=LA_BT_ACTIVE;
            laShowItemFull(uil,cr,Base,p->Identifier,0, 0,laui_ManagedPropInstance,0)->Flags|=LA_UI_FLAGS_NO_DECAL|LA_UI_COLLECTION_NO_HIGHLIGHT;
        }laEndCondition(uil,b);
    }

    laShowSeparator(uil,c);
}
void laui_ManagedProp(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c=laFirstColumn(uil);
    laPropPack PP={0}; if(!Base ||! Base->EndInstance) return;

    laManagedSaveProp* msp=Base->EndInstance;

    laShowLabel(uil,c,msp->Path->Ptr,0,0);
    laShowItemFull(uil,c,0,msp->Path->Ptr,0, 0,laui_ManagedPropInstance,0)->Flags|=LA_UI_FLAGS_NO_DECAL|LA_UI_COLLECTION_NO_HIGHLIGHT;
}
void laui_ManagedSavePanel(laUiList *uil, laPropPack *Base, laPropPack *Operator, laColumn *ExtraColumns, int context){
    laManagedSaveExtra*mse=Operator->EndInstance;
    laColumn *c=laFirstColumn(uil);

    if(mse->OnExit){
        laShowLabel(uil,c,"You still have unsaved/unassigned datablocks:",0,0)->Flags|=LA_TEXT_ALIGN_CENTER|LA_TEXT_LINE_WRAP;
        laShowSeparator(uil,c);
    }

    laUiItem* r=laBeginRow(uil,c,0,0);
    laShowLabel(uil,c,"Viewing",0,0);
    laShowItem(uil,c,Operator,"show_page")->Flags|=LA_UI_FLAGS_EXPAND;
    laEndRow(uil,r);
    laUiItem* g=laMakeEmptyGroup(uil,c,"List",0);{ g->State=LA_UI_ACTIVE;
        laUiList* gu=g->Page; gu->HeightCoeff=mse->OnExit?-4:-3; laColumn* gc=laFirstColumn(gu);
        laUiItem* b=laOnConditionThat(gu,gc,laEqual(laPropExpression(Operator, "show_page"),laIntExpression(0)));{
            laShowItemFull(gu,gc,0,"la.managed_props",0, 0,laui_ManagedProp,0)->Flags|=LA_UI_FLAGS_NO_DECAL|LA_UI_COLLECTION_NO_HIGHLIGHT;
        }laElse(gu,b);{
            laShowItemFull(gu,gc,0,"la.managed_udfs",0,0,0,0)->Flags|=LA_UI_FLAGS_NO_DECAL;
        }laEndCondition(gu,b);     
    }
    
    if(mse->OnExit){
        laShowLabel(uil,c,"If you exit the program now, you will lose those changes.",0,0)->Flags|=LA_TEXT_ALIGN_CENTER|LA_TEXT_LINE_WRAP;
    }

    r=laBeginRow(uil,c,0,0);
    if(mse->OnExit){
        laShowItemFull(uil,c,0,"LA_cancel",0,"text=Discard And Quit;feedback=DISCARD_AND_QUIT",0,0);
    }
    laShowSeparator(uil,c)->Expand=1;
    laShowItemFull(uil,c,0,"LA_managed_save",0,"quiet=true;ignore_unassigned=true;text=Save All Modified;feedback=SAVE_MODIFIED",0,0);
    laEndRow(uil,r);
}

void laui_ResourceFolderItem(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c;

    c = laFirstColumn(uil);
    laUiItem* b=laBeginRow(uil,c,0,0);
    laShowItem(uil, c, Base, "remove")->Flags|=LA_UI_FLAGS_ICON;
    laUiItem* ui=laShowItem(uil, c, Base, "path");
    ui->Expand=1;
    laShowItem(uil,c,&ui->PP,"get_folder_path")->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b);
}

void laui_LogItem(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laShowItem(uil, ExtraColumns, Base, "content")->Flags|=LA_TEXT_LINE_WRAP|LA_TEXT_USE_NEWLINE|LA_TEXT_MONO|LA_UI_FLAGS_NO_TOOLTIP;
}

void laui_NodeRack(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laNodeRack* r=This->EndInstance;
    laColumn* c=laFirstColumn(uil),*cl, *cr;
    laShowLabel(uil,c,"\n",0,0);
    laUiItem* b=laBeginRow(uil,c,0,0);
    laShowItemFull(uil,c,This,"move",0,"icon=🡰;direction=left;",0,0)->Flags|=LA_UI_FLAGS_ICON;
    laUiItem* u=laShowItem(uil,c,This,"name"); u->Flags|=LA_UI_FLAGS_NO_DECAL|LA_TEXT_ALIGN_CENTER; u->Expand=1;
    laShowItemFull(uil,c,This,"move",0,"icon=🡲;direction=right;",0,0)->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b);
    laShowLabel(uil,c,"\n",0,0);
    laShowItem(uil,c,This,"nodes")->Flags|=LA_UI_FLAGS_NO_DECAL;
    laShowLabel(uil,c,"\n",0,0);
    b=laBeginRow(uil,c,0,0);
    laShowItemFull(uil,c,This,"add_node",0,(context==LA_RACK_TYPE_DRIVER)?"target=DRIVER;":"target=INPUT;",0,0)->Expand=1;
    laShowItem(uil,c,This,"delete")->Flags|=LA_UI_FLAGS_ICON;
    laShowItem(uil,c,This,"insert_rack");
    laEndRow(uil,b);
    laShowLabel(uil,c,"\n\n\n",0,0);
}
void laui_RackPage(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn* c=laFirstColumn(uil); laRackPage* r=This->EndInstance;
    laUiItem* b=laOnConditionThat(uil,c,laPropExpression(This,"has_rack"));{
        laUiItem* g=laMakeGroup(uil,c,"Racks",0);{ g->Flags|=/*LA_UI_FLAGS_NO_GAP|LA_UI_FLAGS_NO_DECAL|*/LA_UI_FLAGS_NODE_CONTAINER;
            laUiList* gu=g->Page; laColumn* gc=laFirstColumn(gu); gu->AllowScale=1; gu->HeightCoeff=-1; g->State=LA_UI_ACTIVE;
            laUiItem* hui=laShowItemFull(gu,gc,This,"racks",0,0,laui_NodeRack,r?r->RackType:0); hui->Expand=15; hui->Flags|=LA_UI_FLAGS_NO_DECAL;
        }
    }laElse(uil,b);{
        laShowItemFull(uil,c,This,"add_rack",0,0,0,0);
    }laEndCondition(uil,b);
}
void laui_NodeCategory(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn* c=laFirstColumn(uil); laNodeCategory* nc=This->EndInstance;
    if(nc->Ui){ nc->Ui(uil,This,Extra,0,context); return; }
    char buf[256];
    for(laListItemPointer* lip=nc->NodeTypes.pFirst;lip;lip=lip->pNext){
        laBaseNodeType* nt=lip->p; sprintf(buf,"feedback=%s;text=%s;icon=%lc;",nt->TypeName,nt->Name,nt->Icon?nt->Icon:' ');
        laShowItemFull(uil,c,0,"LA_confirm",0,buf,0,0);
    }
}

void laui_UserPreference(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn* c = laFirstColumn(uil),*cl,*cr; laSplitColumn(uil,c,0.5);cl=laLeftColumn(c,0);cr=laRightColumn(c,0);
    laUiList *muil;
    laColumn *mc, *mcl, *mcr, *mcll, *mclr;
    laUiItem *b, *bracket;

    //_LA_PANEL_USER_PREFERENCE = p;
    //la_UDFAppendSharedTypePointer("_LA_PANEL_USER_PREFERENCE", _LA_PANEL_USER_PREFERENCE);
    //p->Show = 0;
    //p->FastUpdate = 1;

    b=laBeginRow(uil,cl,0,0); laUiItem* ui=laShowLabel(uil,cl,"Save on exit:",0,0); ui->Expand=1; ui->Flags|=LA_TEXT_ALIGN_RIGHT;
    laShowItemFull(uil,cl,0,"la.user_preferences.save_preferences_on_exit",LA_WIDGET_ENUM_HIGHLIGHT,"text=🞆",0,0);
    laEndRow(uil,b);
    laShowItem(uil,cr,0,"LA_save_user_preferences");

    bracket = laMakeTab(uil, c, 0);{

        for(laExtraPreferencePage* epp=MAIN.ExtraPreferencePages.pFirst;epp;epp=epp->Item.pNext){
            muil = laAddTabPage(bracket, epp->Name); epp->Func(muil,0,0,0,0);
        }

        muil = laAddTabPage(bracket, "Display");{
            //muil->HeightCoeff = -1;
            mc = laFirstColumn(muil);
            laSplitColumn(muil, mc, 0.5);
            mcl = laLeftColumn(mc, 0);
            mcr = laRightColumn(mc, 0);

            if(MAIN.PreferencePageDisplay){ MAIN.PreferencePageDisplay(muil,0,0,0,0); }

            laShowLabel(muil, mc, "Interface:", 0, 0);
            laShowLabel(muil, mcl, "Row Height:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.interface_size");
            laShowLabel(muil, mcl, "Font Size:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.font_size");
            laShowLabel(muil, mcl, "Margin Size:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.margin_size");
            laShowLabel(muil, mcl, "Panel Multisample:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.panel_multisample")->Flags|=LA_UI_FLAGS_EXPAND;
            laShowItem(muil, mc, 0, "la.user_preferences.top_framerate");

            laShowSeparator(muil, mc);

            laShowLabel(muil, mc, "Color:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.color_picker_gamma");

            laShowSeparator(muil, mc);

            laShowLabel(muil, mc, "Nodes:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.wire_color_slices");
            laShowItem(muil, mcr, 0, "la.user_preferences.wire_thickness");
            laShowItem(muil, mcr, 0, "la.user_preferences.wire_saggyness");

            laShowSeparator(muil, mc);

            laShowLabel(muil, mc, "Floating Panel:", 0, 0);
            laShowItem(muil, mcl, 0, "la.user_preferences.floating_alpha");
            laShowItem(muil, mcr, 0, "la.user_preferences.solid_shadow_length");

            laShowSeparator(muil, mc);

            laShowLabel(muil, mc, "Translation:", 0, 0);
            laShowLabel(muil, mcl, "Enable Translation:", 0, 0);
            laShowItemFull(muil, mcr, 0, "la.user_preferences.enable_translation",LA_WIDGET_ENUM_CYCLE,0,0,0);
            b = laOnConditionThat(muil, mcl, laPropExpression(0, "la.user_preferences.enable_translation"));{
            laShowLabel(muil, mcl, "Language:", 0, 0);
                laShowItemFull(muil, mcr, 0, "la.user_preferences.languages", LA_WIDGET_COLLECTION_SELECTOR, 0, 0, 0);
                laShowItem(muil, mcl, 0, "LA_translation_dump");
            }
            laEndCondition(muil, b);
        }
        
        muil = laAddTabPage(bracket, "Input");{
            //muil->HeightCoeff = -1;
            mc = laFirstColumn(muil);
            laSplitColumn(muil, mc, 0.5);
            mcl = laLeftColumn(mc, 0);
            mcr = laRightColumn(mc, 0);

            if(MAIN.PreferencePageInput){ MAIN.PreferencePageInput(muil,0,0,0,0); }

            laShowLabel(muil, mc, "User Interactions:", 0, 0);
            laShowItem(muil, mc, 0, "la.user_preferences.scroll_speed");
            laShowItem(muil, mcl, 0, "la.user_preferences.animation_speed");
            laShowItem(muil, mcr, 0, "la.user_preferences.panel_animation_speed");
            laShowItem(muil, mc, 0, "la.user_preferences.valuator_threshold");
            laShowItem(muil, mc, 0, "la.user_preferences.zoom_speed_2d");
            laShowItem(muil, mcl, 0, "la.user_preferences.tooltip_close_distance");
            laShowItem(muil, mcr, 0, "la.user_preferences.idle_time");

            laShowSeparator(muil,mc);

            laShowLabel(muil, mc, "Input device handling:", 0, 0);
            laShowLabel(muil, mcl, "Input Mapping:", 0, 0)->Flags|=LA_TEXT_ALIGN_RIGHT;
            laShowItemFull(muil, mcr, 0, "LA_panel_activator",0,"panel_id=LAUI_input_mapper;text=Configure",0,0);
            laShowLabel(muil, mcl, "Pads/Joysticks:", 0, 0)->Flags|=LA_TEXT_ALIGN_RIGHT;
            laShowItemFull(muil, mcr, 0, "LA_panel_activator",0,"panel_id=LAUI_controllers;text=Configure",0,0);

            laShowSeparator(muil,mc);

            laShowLabel(muil, mc, "Wacom Devices:", 0, 0);
            laShowItem(muil, mcl, 0, "la.user_preferences.wacom_device_stylus");
            laShowItem(muil, mcl, 0, "la.user_preferences.wacom_device_eraser");
            laShowItemFull(muil, mcr, 0, "LA_refresh_controllers",0,"text=Refresh",0,0);
        }

        muil = laAddTabPage(bracket, "Resource");{
            //muil->HeightCoeff = -1;
            mc = laFirstColumn(muil);

            laSplitColumn(muil, mc, 0.5);
            mcl = laLeftColumn(mc, 0);
            mcr = laRightColumn(mc, 0);

            if(MAIN.PreferencePageResource){ MAIN.PreferencePageResource(muil,0,0,0,0); }

            laShowLabel(muil, mcl, "UDF Manager Default View:", 0, 0);
            laShowItem(muil, mcr, 0, "la.user_preferences.manager_default_view")->Flags|=LA_UI_FLAGS_EXPAND;

            laShowSeparator(muil,mc);

            b=laBeginRow(muil,mc,0,0);
            laShowLabel(muil, mc, "Resource Folders", 0, 0)->Expand=1;
            laShowItem(muil, mc, 0, "LA_add_resource_folder");
            laEndRow(muil,b);
            laShowItem(muil, mc, 0, "la.user_preferences.resource_folders");
        }

        muil = laAddTabPage(bracket, "Theme");{
            //muil->HeightCoeff = -1;
            mc = laFirstColumn(muil);

            if(MAIN.PreferencePageTheme){ MAIN.PreferencePageTheme(muil,0,0,0,0); }

            laShowLabel(muil, mc, "Active Theme", 0, 0);
            laShowItemFull(muil, mc, 0, "la.themes", LA_WIDGET_COLLECTION_SELECTOR, 0,laui_ThemeListItem,0);
            laShowItemFull(muil, mc, 0, "la.themes", LA_WIDGET_COLLECTION_SINGLE, 0 ,0,0);
        }
    }
}
void laui_About(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn *c, *cl, *cr, *tc, *gc, *gcl, *gcr, *tcl, *tcr;
    laUiItem *bracket, *t, *g;
    laPanel *p;
    laUiList *u, *gu;
    laPropContainer *pc;
    int h;
    laOperatorType *at;
    laUiList *first;
    char buf[2048]={0};

    c = laFirstColumn(uil);

    if(MAIN.AboutContent){
        MAIN.AboutContent(uil,Base,OperatorInst,ExtraColumns,0);
    }else{
        laShowLabel(uil, c, "LaGUI", 0, 0);
        laShowLabel(uil, c, "A graphical user interface application toolkit", 0, 0)->Flags|=LA_TEXT_LINE_WRAP;
        laShowLabel(uil, c, "(C)Yiming Wu", 0, 0);
    }

    t = laMakeTab(uil, c, 0);{
        first = u = laAddTabPage(t, "Version");
        tc = laFirstColumn(u);

        if(MAIN.AboutVersion){
            MAIN.AboutVersion(u,Base,OperatorInst,ExtraColumns,0);
        }

        g = laMakeGroup(u, tc, "LaGUI information", 0);
        gu = g->Page;{
            gc = laFirstColumn(gu);
            sprintf(buf, "LaGUI.%d.%d.%d", LA_VERSION_MAIN, LA_VERSION_SUB, LA_VERSION_MINOR);
            laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            laShowLabel(gu, gc, LAGUI_GIT_BRANCH,0,0)->Flags|=LA_TEXT_MONO;
#ifdef LAGUI_GIT_HASH
            laShowLabel(gu, gc, LAGUI_GIT_HASH,0,0)->Flags|=LA_TEXT_MONO;
#endif
            sprintf(buf, "Built on:[UTC+8] %s %s", __DATE__, __TIME__);
            laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            sprintf(buf, "UDF Identifier: %s",LA_UDF_IDENTIFIER);
            laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            laShowLabel(gu, gc, "UDF Extensions:", 0, 0)->Flags|=LA_TEXT_MONO;
            laShowLabel(gu, gc, "🗸LA_UDF_BASICS", 0, 0)->Flags|=LA_TEXT_MONO;
            for(int i=0;i<64;i++){
                if(!LA_UDF_EXTENSION_STRINGS[i][0]) break; int gray=!((1<<i)&LA_UDF_EXTENSION_BITS);
                sprintf(buf,"%s%s",gray?"  ":"🗸",LA_UDF_EXTENSION_STRINGS[i]);
                laUiItem* ext=laShowLabel(gu, gc, buf, 0, 0); ext->Flags|=LA_TEXT_MONO;
                if(gray){ ext->Flags|=LA_UI_FLAGS_DISABLED; }
            }

        }
        
        g = laMakeGroup(u, tc, "Device Graphics", 0);
        gu = g->Page;{
            gc = laFirstColumn(gu);
            sprintf(buf, "OpenGL Version: %s", T->GLVersionStr);   laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            sprintf(buf, "OpenGL Vendor: %s", T->GLVendorStr);     laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            sprintf(buf, "OpenGL Renderer: %s", T->GLRendererStr); laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
            sprintf(buf, "GLSL Version: %s", T->GLSLVersionStr);   laShowLabel(gu, gc, buf, 0, 0)->Flags|=LA_TEXT_MONO;
        }

        u = laAddTabPage(t, "Authors");
        tc= laFirstColumn(u);

        if(MAIN.AboutAuthor){
            MAIN.AboutAuthor(u,Base,OperatorInst,ExtraColumns,0);
        }

        g = laMakeGroup(u, tc, "LaGUI", 0);
        gu = g->Page;{
            gc = laFirstColumn(gu);
            laShowLabel(gu,gc,"LaGUI application framework is made by Wu Yiming.",0,0)->Flags|=LA_TEXT_LINE_WRAP;
            laUiItem* b=laBeginRow(gu,gc,0,0);
            laShowItemFull(gu, gc, 0, "LA_open_internet_link", 0, "link=http://www.ChengduLittleA.com/lagui;text=Details", 0, 0);
            laShowItemFull(gu, gc, 0, "LA_open_internet_link", 0, "link=http://www.ChengduLittleA.com;text=Yiming's Blog", 0, 0);
            laEndRow(gu,b);
        }

        t->Page = first;
    }
}
void laui_UndoHistories(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn* c=laFirstColumn(uil),*cl,*cr; laSplitColumn(uil,c,0.5);cl=laLeftColumn(c,0);cr=laRightColumn(c,0);
    laShowItem(uil,cl,0,"LA_undo"); laShowItem(uil,cr,0,"LA_redo")->Flags|=LA_TEXT_ALIGN_RIGHT;
    laUiItem* g=laMakeGroup(uil,c,"123",0);{ g->State=LA_UI_ACTIVE; g->Flags|=LA_UI_FLAGS_PREFER_BOTTOM;
        laUiList* gu=g->Page; laColumn* gc=laFirstColumn(gu); gu->HeightCoeff=-1;
        laShowItem(gu,gc,0,"la.differences")->Flags|=LA_UI_FLAGS_NO_DECAL;
    }
}
void laui_Terminal(laUiList *uil, laPropPack *Base, laPropPack *OperatorInst, laColumn *ExtraColumns, int context){
    laColumn* c=laFirstColumn(uil);
    laUiItem* g=laMakeGroup(uil,c,"123",0);{ g->State=LA_UI_ACTIVE; g->Flags|=LA_UI_FLAGS_PREFER_BOTTOM;
        laUiList* gu=g->Page; laColumn* gc=laFirstColumn(gu); gu->HeightCoeff=-1;
        laShowItem(gu,gc,0,"la.logs")->Flags|=LA_UI_FLAGS_NO_DECAL;
    }
}
void laui_IdleDataManager(laUiList *uil, laPropPack *Base, laPropPack *Extra, laColumn *ExtraColumns, int context){
    laColumn *c=laFirstColumn(uil);

    laUiItem* r=laBeginRow(uil,c,0,0);
    laShowLabel(uil,c,"Viewing",0,0);
    laShowItem(uil,c,Extra,"show_page")->Flags|=LA_UI_FLAGS_EXPAND;
    laEndRow(uil,r);
    laUiItem* g=laMakeEmptyGroup(uil,c,"List",0);{ g->State=LA_UI_ACTIVE;
        laUiList* gu=g->Page; gu->HeightCoeff=-3; laColumn* gc=laFirstColumn(gu);
        laUiItem* b=laOnConditionThat(gu,gc,laEqual(laPropExpression(Extra, "show_page"),laIntExpression(0)));{
            laShowItemFull(gu,gc,0,"la.managed_props",0, 0,laui_ManagedProp,0)->Flags|=LA_UI_FLAGS_NO_DECAL;
        }laElse(gu,b);{
            laShowItemFull(gu,gc,0,"la.managed_udfs",0,0,0,0)->Flags|=LA_UI_FLAGS_NO_DECAL;
        }laEndCondition(gu,b);     
    }
    
    r=laBeginRow(uil,c,0,0);
    laShowSeparator(uil,c)->Expand=1;
    laShowItemFull(uil,c,0,"LA_managed_save",0,"quiet=true;ignore_unassigned=true;text=Save All Modified",0,0);
    laEndRow(uil,r);
}
void lauidetached_IdleDataManager(laPanel* p){
    la_MakeDetachedProp(p, "la.user_preferences.manager_default_view", "show_page");
}
void laui_TextureInspector(laUiList *uil, laPropPack *Base, laPropPack *Extra, laColumn *ExtraColumns, int context){
    laColumn *c=laFirstColumn(uil),*cl,*cr;
    laSplitColumn(0, c, 0.5);
    cl = laLeftColumn(c, 0);
    cr = laRightColumn(c, 0);

    laShowLabel(uil,cl,"Viewing Texture:",0,0)->Flags|=LA_TEXT_ALIGN_RIGHT;
    laShowItemFull(uil,cr,Extra,"textures",LA_WIDGET_COLLECTION_SELECTOR,0,0,0);
    laUiItem* u2=laShowCanvas(uil,c,Extra,"textures",0,-1);
    laDefault2DViewOverlayRight(u2);
}
void lauidetached_TextureInspector(laPanel* p){
    la_MakeDetachedProp(p, "tns.texture_list", "textures");
}
void laui_GameController(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn* c=laFirstColumn(uil),*cl, *cr;
    laSplitColumn(uil,c,0.5); cl=laLeftColumn(c,0); cr=laRightColumn(c,0);
    
    laUiItem* b=laBeginRow(uil,cl,0,0);
    laUiItem* l=laShowLabel(uil,cl,"Controller:",0,0); l->Flags|=LA_TEXT_ALIGN_RIGHT; l->Expand=1;
    laShowItem(uil,cl,0,"LA_refresh_controllers")->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b);

    laShowItemFull(uil,cr,Extra,"controllers",LA_WIDGET_COLLECTION_SELECTOR,0,laui_IdentifierOnly,0);

    b=laOnConditionThat(uil,c,laPropExpression(Extra,"controllers"));{
        laShowItemFull(uil,c,Extra,"controllers",LA_WIDGET_COLLECTION_SINGLE,0,0,0);
    }laElse(uil,b);{
        laShowLabel(uil,c,"Please select a controller.",0,0)->Flags|=LA_TEXT_ALIGN_CENTER;
    }
}
void lauidetached_GameController(laPanel* p){
    la_MakeDetachedProp(p, "la.controllers", "controllers");
}
void laui_InputMapper(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn* c=laFirstColumn(uil),*cl, *cr;
    laSplitColumn(uil,c,0.6); cl=laLeftColumn(c,0); cr=laRightColumn(c,0);
    
    laUiItem* b=laBeginRow(uil,cl,0,0);
    laUiItem* b2=laOnConditionThat(uil,cl,laPropExpression(0,"la.input_mapping.pages"));{
        laUiItem* b3=laOnConditionThat(uil,cl,laPropExpression(0,"la.input_mapping.current_page"));{
            laShowItem(uil,cr,0,"la.input_mapping.current_page.name")->Expand=1;
            laShowItem(uil,c,Extra,"LA_input_mapping_rebuild")->Flags|=LA_UI_FLAGS_ICON;
        }laEndCondition(uil,b3);
    }laEndCondition(uil,b2);
    laShowItem(uil,cl,0,"LA_add_input_mapping_page")->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b);
    b2=laOnConditionThat(uil,cr,laPropExpression(0,"la.input_mapping.pages"));{
        laShowItemFull(uil,cr,0,"la.input_mapping.pages",LA_WIDGET_COLLECTION_SELECTOR,0,0,0);
    }laEndCondition(uil,b2);
    
    laShowItemFull(uil,c,0,"la.input_mapping.current_page",LA_WIDGET_COLLECTION_SINGLE,0,laui_RackPage,0)->Flags|=LA_UI_FLAGS_NO_DECAL;;
}
void lauidetached_Drivers(laPanel* p){
    la_MakeDetachedProp(p, "la.drivers.current_page", "page");
}
void laui_Drivers(laUiList *uil, laPropPack *This, laPropPack *Extra, laColumn *UNUSED, int context){
    laColumn* c=laFirstColumn(uil),*cl, *cr;
    laSplitColumn(uil,c,0.6); cl=laLeftColumn(c,0); cr=laRightColumn(c,0);
    
    laUiItem* b=laBeginRow(uil,cl,0,0);
    laUiItem* b2=laOnConditionThat(uil,cl,laPropExpression(0,"la.drivers.pages"));{
        laUiItem* b3=laOnConditionThat(uil,cl,laPropExpression(Extra,"page"));{
            laShowItem(uil,cr,Extra,"page.name")->Expand=1;
            laShowItem(uil,c,Extra,"LA_driver_rebuild")->Flags|=LA_UI_FLAGS_ICON;
        }laEndCondition(uil,b3);
    }laEndCondition(uil,b2);
    laShowItem(uil,cl,0,"LA_add_driver_page")->Flags|=LA_UI_FLAGS_ICON;
    laEndRow(uil,b);
    b2=laOnConditionThat(uil,cr,laPropExpression(0,"la.drivers.pages"));{
        laShowItemFull(uil,cr,Extra,"page",LA_WIDGET_COLLECTION_SELECTOR,0,0,0);
    }laEndCondition(uil,b2);
    
    laShowItemFull(uil,c,Extra,"page",LA_WIDGET_COLLECTION_SINGLE,0,laui_RackPage,0)->Flags|=LA_UI_FLAGS_NO_DECAL;;
}


void la_RegisterBuiltinTemplates(){
    laRegisterUiTemplate("LAUI_input_mapper","Input Mapper",laui_InputMapper,0,0,"Controlling",0,0,0);
    laRegisterUiTemplate("LAUI_drivers","Drivers",laui_Drivers,lauidetached_Drivers,0,0,0,0,0);
    laRegisterUiTemplate("LAUI_controllers", "Controllers", laui_GameController, lauidetached_GameController, 0,0,0,0,0);
    laRegisterUiTemplate("LAUI_user_preferences", "User Preferences", laui_UserPreference, 0, 0, "System",0,25,25);
    laRegisterUiTemplate("LAUI_about", "About", laui_About, 0, 0, 0, 0,15,25);
    laRegisterUiTemplate("LAUI_texture_inspector", "Texture Inspector", laui_TextureInspector, lauidetached_TextureInspector, 0, 0, 0,0,0);
    laRegisterUiTemplate("LAUI_data_manager", "Data Manager", laui_IdleDataManager, lauidetached_IdleDataManager, 0, 0, 0,25,25);
    laRegisterUiTemplate("LAUI_histories", "Histories", laui_UndoHistories, 0, 0, 0, 0,10,25);
    laRegisterUiTemplate("LAUI_terminal", "Terminal", laui_Terminal, 0, 0, 0, 0,20,25);
}
